package operators

import (
	"bufio"
	"encoding/json"
	"fmt"
	"os"
	"path/filepath"
	"regexp"
	"strings"

	git "gopkg.in/src-d/go-git.v4"
)

type genOpts struct {
	Title        string   `json:"title"`
	Description  string   `json:"description"`
	Author       string   `json:"author"`
	Version      string   `json:"version"`
	Homepage     string   `json:"homepage"`
	Repository   string   `json:"repository"`
	Tag          string   `json:"tag"`
	Dependencies []string `json:"dependencies"`
}

func (g genOpts) String() string {
	var out strings.Builder
	out.WriteString(fmt.Sprintf("\033[33;1m%s\033[0m\n", g.Title))
	out.WriteString(fmt.Sprintf("  \033[2m├\033[0m Description: %s\n", g.Description))
	out.WriteString(fmt.Sprintf("  \033[2m├\033[0m Author: %s\n", g.Author))
	out.WriteString(fmt.Sprintf("  \033[2m├\033[0m Version: %s\n", g.Version))
	out.WriteString(fmt.Sprintf("  \033[2m├\033[0m Repository: %s\n", g.Repository))
	out.WriteString(fmt.Sprintf("  \033[2m├\033[0m Tag: %s\n", g.Tag))
	out.WriteString(fmt.Sprintf("  \033[2m├\033[0m Homepage: %s\n", g.Homepage))
	if len(g.Dependencies) == 0 {
		out.WriteString("  \033[2m└\033[0m Dependencies: \033[3mno dependencies\033[0m\n")
	} else {
		out.WriteString(fmt.Sprintf("  \033[2m└\033[0m Dependencies: %s\n", ""))
		for i, u := range g.Dependencies {
			if i < len(g.Dependencies)-1 {
				out.WriteString(fmt.Sprintf("    \033[2m├\033[0m %s\n", u))
			} else {
				out.WriteString(fmt.Sprintf("    \033[2m└\033[0m %s\n", u))
			}
		}

	}
	if len(g.Dependencies) == 0 {
	}
	return out.String()
}

// Generate generates a package
// by cloning the package template repo
func Generate(solo bool) error {
	opts, err := promptOptions()
	if err != nil {
		return err
	}

	j, err := json.MarshalIndent(*opts, "", "\t")
	if err != nil {
		return err
	}

	pwd, err := os.Getwd()
	if err != nil {
		return err
	}

	dir := filepath.Join(pwd, opts.Title)

	if solo {
		err = os.MkdirAll(filepath.Join(dir, ".solo", "tree", "master"), 0755)
		if err != nil {
			return err
		}
		err = os.MkdirAll(filepath.Join(dir, ".solo", "tags"), 0755)
		if err != nil {
			return err
		}
		f, err := os.Create(filepath.Join(dir, ".solo", "solo.conf"))
		if err != nil {
			return err
		}
		defer f.Close()
		host, _ := os.Hostname()
		f.WriteString(fmt.Sprintf(`(("heads" ()) ("branch" "master") ("tags" ()) ("description" "%s") ("author" "%s@%s") ("version" (0 3 0)) ("remote" #f) ("floating?" #f))`, opts.Description, opts.Author, host))
	} else {
		_, err = git.PlainInit(dir, false)
		if err != nil {
			return err
		}
	}

	jf, err := os.Create(filepath.Join(dir, "module.json"))
	if err != nil {
		return err
	}
	defer jf.Close()
	jf.Write(j)

	sf, err := os.Create(filepath.Join(dir, "main.slo"))
	if err != nil {
		return err
	}
	defer sf.Close()
	sf.WriteString(";;;\n")
	sf.WriteString(fmt.Sprintf(";;; %s - %s\n", opts.Title, opts.Description))
	sf.WriteString(fmt.Sprintf(";;; Author: %s\n", opts.Author))
	sf.WriteString(fmt.Sprintf(";;; Version: %s\n", opts.Version))
	sf.WriteString(";;;\n\n; display \"Hello, World!\"\n")
	sf.WriteString("(define hello-world (lambda ()\n")
	sf.WriteString(fmt.Sprintf("  (display \"'Hello, world!' from %s\\n\")))\n\n; vim: ts=2 sw=2 expandtab ft=slope\n", opts.Title))

	rm, err := os.Create(filepath.Join(dir, "README.md"))
	if err != nil {
		return err
	}
	defer rm.Close()
	rm.WriteString("# ")
	rm.WriteString(opts.Title)
	rm.WriteString("\n\n")
	rm.WriteString(opts.Description)
	rm.WriteString("\n\n")

	fmt.Printf("\nPackage \033[1m%q\033[0m has been created!\n", opts.Title)

	return nil
}

func promptOptions() (*genOpts, error) {
	reader := bufio.NewReader(os.Stdin)

	fmt.Println("Mandatory fields are marked with an (*).\nNon-mandatory fields may be left blank.\nNote that fields that are required by the package registry are\nnot mandatory here simply due to the fact that the information\nmay not be available when starting a new project.\nRemember to go back and update this information.\n\n", "")

	// A title is entirely composed of lowercase letters, digits, dashes, and underscores
	title, err := promptOption(
		"package title*",
		`^[a-z\d-_]+$`,
		"the title can only contain lowercase letters, digits, dashes, and underscores",
		reader,
	)

	if err != nil {
		return nil, err
	}

	description, err := promptOption(
		"description*",
		`^.{1,60}$`,
		"the description must be at least 1 character and less than 60 characters",
		reader,
	)
	if err != nil {
		return nil, err
	}

	author, err := promptOption(
		"author*",
		`^.+$`,
		"the author's name must be at least 1 character",
		reader,
	)
	if err != nil {
		return nil, err
	}

	version, err := promptOption(
		"version",
		`.*`,
		"",
		reader,
	)
	if err != nil {
		return nil, err
	}

	repo, err := promptOption(
		"repository",
		`.*`,
		"",
		reader,
	)
	if err != nil {
		return nil, err
	}

	tag, err := promptOption(
		"tag",
		`.*`,
		"",
		reader,
	)
	if err != nil {
		return nil, err
	}

	homepage, err := promptOption(
		"homepage",
		`.*`,
		"",
		reader,
	)
	if err != nil {
		return nil, err
	}

	return &genOpts{
		Title:        title,
		Description:  description,
		Author:       author,
		Version:      version,
		Repository:   repo,
		Homepage:     homepage,
		Tag:          tag,
		Dependencies: make([]string, 0),
	}, nil
}

func promptOption(msg, pattern, errMsg string, reader *bufio.Reader) (string, error) {
	var (
		text string
		err  error
	)

	for {
		fmt.Printf("\033[1m%15s:\033[0m ", msg)

		text, err = reader.ReadString('\n')
		if err != nil {
			return "", err
		}

		text = strings.TrimSpace(text)

		matches, err := regexp.MatchString(pattern, text)
		if err != nil {
			return "", err
		}

		if matches {
			break
		}

		fmt.Printf("\033[31m  invalid input:\033[0m %s\n", errMsg)
	}

	return text, nil
}
