package operators

import (
	"encoding/json"
	"fmt"
	"io"
	"io/ioutil"
	"net/http"
	"os"
	"os/user"
	"path/filepath"
	"strings"
	"time"
)

const (
	packageListURL string = "https://git.rawtext.club/slope-lang/packages/raw/branch/master/packages.json"
	GlobalPath     string = "/usr/local/lib/slope/modules/"
)

func ExpandedAbsFilepath(p string) string {
	if strings.HasPrefix(p, "~") {
		if p == "~" || strings.HasPrefix(p, "~/") {
			homedir, _ := os.UserHomeDir()
			if len(p) <= 2 {
				p = homedir
			} else if len(p) > 2 {
				p = filepath.Join(homedir, p[2:])
			}
		} else {
			i := strings.IndexRune(p, '/')
			var u string
			var remainder string
			if i < 0 {
				u = p[1:]
				remainder = ""
			} else {
				u = p[1:i]
				remainder = p[i:]
			}
			usr, err := user.Lookup(u)
			if err != nil {
				p = filepath.Join("/home", u, remainder)
			} else {
				p = filepath.Join(usr.HomeDir, remainder)
			}
		}
	} else if !strings.HasPrefix(p, "/") {
		wd, _ := os.Getwd()
		p = filepath.Join(wd, p)
	}

	path, _ := filepath.Abs(p)
	return path
}

func GetModBaseDir() string {
	p := os.Getenv("SLOPE_MOD_PATH")
	if p == "" {
		x := os.Getenv("XDG_DATA_HOME")
		if x == "" {
			return ExpandedAbsFilepath("~/.local/share/slope/modules/")
		}
		return filepath.Join(x, "slope", "modules")
	}
	return p
}

func checkRegistry() error {
	modDir := GetModBaseDir()
	path := filepath.Join(modDir, "package.json")

	f, err := os.Stat(path)
	if err != nil {
		return fmt.Errorf("The registry is not present on the system. Run: `slp fetch`")
	} else if f.ModTime().Before(time.Now().Add(-7 * 24 * time.Hour)) {
		fmt.Fprint(os.Stderr, "You have not run `slp fetch` in more than 7 days: consider updating the package list\n")
		return nil
	}
	return nil
}

func downloadTheDirectory() error {
	modDir := GetModBaseDir()
	path := filepath.Join(modDir, "package.json")

	resp, err := http.Get(packageListURL)
	if err != nil {
		return fmt.Errorf("Unable to retrieve package registry")
	}
	defer resp.Body.Close()

	out, err := os.Create(path)
	if err != nil {
		return err
	}
	defer out.Close()

	_, err = io.Copy(out, resp.Body)
	if err != nil {
		return err
	}

	return nil
}

func getPackages() (map[string]genOpts, error) {
	modDir := GetModBaseDir()

	data, err := ioutil.ReadFile(filepath.Join(modDir, "package.json"))
	if err != nil {
		return nil, err
	}

	var packages map[string]genOpts

	err = json.Unmarshal(data, &packages)
	if err != nil {
		return nil, err
	}

	return packages, nil
}
